@@ -19,8 +19,8 @@ module Numeric.Montgomery.Secp256k1.Curve (
   -- * Montgomery form, secp256k1 field prime modulus
     Montgomery(..)
   , render
-  , to
-  , from
+  , to_vartime
+  , from_vartime
   , zero
   , one
 
@@ -51,11 +51,12 @@ module Numeric.Montgomery.Secp256k1.Curve (
   , neg#
   , inv
   , inv#
-  , sqrt
+  , sqrt_vartime
   , sqrt#
   , exp
+  , exp#
   , odd#
-  , odd
+  , odd_vartime
   ) where
 
 import Control.DeepSeq
@@ -66,13 +67,13 @@ import qualified Data.Word.Wide as W
 import Data.Word.Wider (Wider(..))
 import qualified Data.Word.Wider as WW
 import GHC.Exts (Word(..))
-import Prelude hiding (or, and, not, sqrt, exp, odd)
+import Prelude hiding (or, and, not, sqrt, exp)
 
 -- montgomery arithmetic, specialized to the secp256k1 field prime modulus
 -- 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F
 
 -- | Montgomery-form 'Wider' words, on the Montgomery domain defined by
---   the secp256k1 scalar group order.
+--   the secp256k1 field prime.
 --
 --   >>> let one = 1 :: Montgomery
 --   >>> one
@@ -92,21 +93,23 @@ render (Montgomery (# Limb a, Limb b, Limb c, Limb d #)) =
   <> show (W# c) <> ", " <> show (W# d) <> ")"
 
 instance Show Montgomery where
-  show = show . from
+  show = show . from_vartime
 
+-- | Note that 'fromInteger' necessarily runs in variable time due
+--   to conversion from the variable-size, potentially heap-allocated
+--   'Integer' type.
 instance Num Montgomery where
   a + b = add a b
   a - b = sub a b
   a * b = mul a b
   negate a = neg a
   abs = id
-  fromInteger = to . WW.to
-  signum a = case a of
-    Montgomery (# Limb 0##, Limb 0##, Limb 0##, Limb 0## #) -> 0
-    _ -> 1
-
-instance Eq Montgomery where
-  a == b = C.decide (eq a b)
+  fromInteger = to_vartime . WW.to_vartime
+  signum (Montgomery (# l0, l1, l2, l3 #)) =
+    let !(Limb l) = l0 `L.or#` l1 `L.or#` l2 `L.or#` l3
+        !n = C.from_word_nonzero# l
+        !b = C.to_word# n
+    in  Montgomery (# Limb b, Limb 0##, Limb 0##, Limb 0## #)
 
 instance NFData Montgomery where
   rnf (Montgomery a) = case a of (# _, _, _, _ #) -> ()
@@ -144,8 +147,8 @@ eq_vartime (Montgomery (Wider -> a)) (Montgomery (Wider -> b)) =
 -- innards --------------------------------------------------------------------
 
 redc_inner#
-  :: (# Limb, Limb, Limb, Limb #)              -- ^ upper limbs
-  -> (# Limb, Limb, Limb, Limb #)              -- ^ lower limbs
+  :: (# Limb, Limb, Limb, Limb #)             -- ^ upper limbs
+  -> (# Limb, Limb, Limb, Limb #)             -- ^ lower limbs
   -> (# (# Limb, Limb, Limb, Limb #), Limb #) -- ^ upper limbs, meta-carry
 redc_inner# (# u0, u1, u2, u3 #) (# l0, l1, l2, l3 #) =
   let !(# m0, m1, m2, m3 #) =
@@ -202,7 +205,7 @@ redc
   -> Montgomery -- ^ reduced value
 redc (Montgomery l) (Montgomery u) =
   let !res = redc# l u
-  in  (Montgomery res)
+  in  Montgomery res
 
 retr_inner#
   :: (# Limb, Limb, Limb, Limb #) -- ^ value in montgomery form
@@ -366,14 +369,14 @@ to# x =
 {-# INLINE to# #-}
 
 -- | Convert a 'Wider' word to the Montgomery domain.
-to :: Wider -> Montgomery
-to (Wider x) = Montgomery (to# x)
+to_vartime :: Wider -> Montgomery
+to_vartime (Wider x) = Montgomery (to# x)
 
 -- | Retrieve a 'Montgomery' word from the Montgomery domain.
 --
 --   This function is a synonym for 'retr'.
-from :: Montgomery -> Wider
-from = retr
+from_vartime :: Montgomery -> Wider
+from_vartime = retr
 
 add#
   :: (# Limb, Limb, Limb, Limb #) -- ^ augend
@@ -987,24 +990,27 @@ inv (Montgomery w) = Montgomery (inv# w)
 
 -- | Square root (Tonelli-Shanks) in the Montgomery domain.
 --
---   For a, return x such that a = x x mod p. Returns nothing if no such
---   square root exists.
+--   Returns 'Nothing' if the square root doesn't exist.
 --
---   >>> sqrt 4
+--   Note that the square root calculation itself is performed in
+--   constant time; we branch only when casting to 'Maybe' at the end.
+--
+--   >>> sqrt_vartime 4
 --   Just 2
---   >>> sqrt 15
+--   >>> sqrt_vartime 15
 --   Just 69211104694897500952317515077652022726490027694212560352756646854116994689233
---   >>> (*) <$> sqrt 15 <*> sqrt 15
+--   >>> (*) <$> sqrt_vartime 15 <*> sqrt_vartime 15
 --   Just 15
-sqrt :: Montgomery -> Maybe Montgomery
-sqrt (Montgomery n) = case sqrt# n of
-  (# a | #) -> Just $! Montgomery a
-  _         -> Nothing
+sqrt_vartime :: Montgomery -> Maybe Montgomery
+sqrt_vartime (Montgomery n) = case sqrt# n of
+  (# a, c #)
+    | C.decide c -> Just $! Montgomery a
+    | otherwise  -> Nothing
 
 -- generated by etc/generate_sqrt.sh
 sqrt#
   :: (# Limb, Limb, Limb, Limb #)
-  -> (# (# Limb, Limb, Limb, Limb #) | () #)
+  -> (# (# Limb, Limb, Limb, Limb #), C.Choice #)
 sqrt# a =
   let !t0 = (# Limb 0x1000003D1##, Limb 0##, Limb 0##, Limb 0## #)
       !t1 = sqr# t0
@@ -1511,9 +1517,7 @@ sqrt# a =
       !t502 = sqr# t501
       !t503 = sqr# t502
       !r = t503
-  in  if   C.decide (WW.eq# (sqr# r) a)
-      then (# r | #)
-      else (# | () #)
+  in  (# r, WW.eq# (sqr# r) a #)
 {-# INLINE sqrt# #-}
 
 -- | Exponentiation in the Montgomery domain.
@@ -1523,31 +1527,42 @@ sqrt# a =
 --   >>> exp 2 10
 --   1024
 exp :: Montgomery -> Wider -> Montgomery
-exp (Montgomery b) (Wider e) =
-  let !one# = (# Limb 0x1000003D1##, Limb 0##, Limb 0##, Limb 0## #)
-      loop !r !_ !_ 0 = r
-      loop !r !m !ex !n =
-        let !(# ne, bit #) = WW.shr1_c# ex
-            !candidate = mul# r m
-            !nr = select# r candidate bit
-            !nm = sqr# m
-        in  loop nr nm ne (n - 1)
-  in  Montgomery (loop one# b e (256 :: Word))
+exp (Montgomery b) (Wider e) = Montgomery (exp# b e)
+
+exp#
+  :: (# Limb, Limb, Limb, Limb #)
+  -> (# Limb, Limb, Limb, Limb #)
+  -> (# Limb, Limb, Limb, Limb #)
+exp# b e =
+  let !o = (# Limb 0x1000003D1##, Limb 0##, Limb 0##, Limb 0## #)
+      loop !r !m !ex n = case n of
+        0 -> r
+        _ ->
+          let !(# ne, bit #) = WW.shr1_c# ex
+              !candidate = mul# r m
+              !nr = select# r candidate bit
+              !nm = sqr# m
+          in  loop nr nm ne (n - 1)
+  in  loop o b e (256 :: Word)
+{-# INLINE exp# #-}
 
 odd# :: (# Limb, Limb, Limb, Limb #) -> C.Choice
 odd# = WW.odd#
-{-# INLINE odd #-}
+{-# INLINE odd# #-}
 
 -- | Check if a 'Montgomery' value is odd.
 --
+--   Note that the comparison is performed in constant time, but we
+--   branch when converting to 'Bool'.
+--
 --   >>> odd 1
 --   True
 --   >>> odd 2
 --   False
 --   >>> Data.Word.Wider.odd (retr 3) -- parity is preserved
 --   True
-odd :: Montgomery -> Bool
-odd (Montgomery m) = C.decide (odd# m)
+odd_vartime :: Montgomery -> Bool
+odd_vartime (Montgomery m) = C.decide (odd# m)
 
 -- constant-time selection ----------------------------------------------------
 
