@@ -19,8 +19,8 @@ module Numeric.Montgomery.Secp256k1.Scalar (
   -- * Montgomery form, secp256k1 scalar group order modulus
     Montgomery(..)
   , render
-  , to
-  , from
+  , to_vartime
+  , from_vartime
   , zero
   , one
 
@@ -52,8 +52,9 @@ module Numeric.Montgomery.Secp256k1.Scalar (
   , inv
   , inv#
   , exp
+  , exp#
   , odd#
-  , odd
+  , odd_vartime
   ) where
 
 import Control.DeepSeq
@@ -64,7 +65,7 @@ import qualified Data.Word.Wide as W
 import Data.Word.Wider (Wider(..))
 import qualified Data.Word.Wider as WW
 import GHC.Exts (Word(..))
-import Prelude hiding (or, and, not, exp, odd)
+import Prelude hiding (or, and, not, exp)
 
 -- montgomery arithmetic, specialized to the secp256k1 scalar group order
 -- 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141
@@ -80,7 +81,7 @@ import Prelude hiding (or, and, not, exp, odd)
 data Montgomery = Montgomery !(# Limb, Limb, Limb, Limb #)
 
 instance Show Montgomery where
-  show = show . from
+  show = show . from_vartime
 
 -- | Render a 'Montgomery' value as a 'String', showing its individual
 --   'Limb's.
@@ -92,19 +93,21 @@ render (Montgomery (# Limb a, Limb b, Limb c, Limb d #)) =
      "(" <> show (W# a) <> ", " <> show (W# b) <> ", "
   <> show (W# c) <> ", " <> show (W# d) <> ")"
 
+-- | Note that 'fromInteger' necessarily runs in variable time due
+--   to conversion from the variable-size, potentially heap-allocated
+--   'Integer' type.
 instance Num Montgomery where
   a + b = add a b
   a - b = sub a b
   a * b = mul a b
   negate a = neg a
   abs = id
-  fromInteger = to . WW.to
-  signum a = case a of
-    Montgomery (# Limb 0##, Limb 0##, Limb 0##, Limb 0## #) -> 0
-    _ -> 1
-
-instance Eq Montgomery where
-  a == b = C.decide (eq a b)
+  fromInteger = to_vartime . WW.to_vartime
+  signum (Montgomery (# l0, l1, l2, l3 #)) =
+    let !(Limb l) = l0 `L.or#` l1 `L.or#` l2 `L.or#` l3
+        !n = C.from_word_nonzero# l
+        !b = C.to_word# n
+    in  Montgomery (# Limb b, Limb 0##, Limb 0##, Limb 0## #)
 
 instance NFData Montgomery where
   rnf (Montgomery a) = case a of (# _, _, _, _ #) -> ()
@@ -364,14 +367,14 @@ to# x =
 {-# INLINE to# #-}
 
 -- | Convert a 'Wider' word to the Montgomery domain.
-to :: Wider -> Montgomery
-to (Wider x) = Montgomery (to# x)
+to_vartime :: Wider -> Montgomery
+to_vartime (Wider x) = Montgomery (to# x)
 
 -- | Retrieve a 'Montgomery' word from the Montgomery domain.
 --
 --   This function is a synonym for 'retr'.
-from :: Montgomery -> Wider
-from = retr
+from_vartime :: Montgomery -> Wider
+from_vartime = retr
 
 add#
   :: (# Limb, Limb, Limb, Limb #) -- ^ augend
@@ -948,32 +951,43 @@ inv (Montgomery w) = Montgomery (inv# w)
 --   >>> exp 2 10
 --   1024
 exp :: Montgomery -> Wider -> Montgomery
-exp (Montgomery b) (Wider e) =
-  let !one# = (# Limb 0x402DA1732FC9BEBF##, Limb 0x4551231950B75FC4##
-              ,  Limb 0x0000000000000001##, Limb 0x0000000000000000## #)
-      loop !r !_ !_ 0 = r
-      loop !r !m !ex !n =
-        let !(# ne, bit #) = WW.shr1_c# ex
-            !candidate = mul# r m
-            !nr = select# r candidate bit
-            !nm = sqr# m
-        in  loop nr nm ne (n - 1)
-  in  Montgomery (loop one# b e (256 :: Word))
+exp (Montgomery b) (Wider e) = Montgomery (exp# b e)
+
+exp#
+  :: (# Limb, Limb, Limb, Limb #)
+  -> (# Limb, Limb, Limb, Limb #)
+  -> (# Limb, Limb, Limb, Limb #)
+exp# b e =
+  let !o = (# Limb 0x402DA1732FC9BEBF##, Limb 0x4551231950B75FC4##
+           ,  Limb 0x0000000000000001##, Limb 0x0000000000000000## #)
+      loop !r !m !ex n = case n of
+        0 -> r
+        _ ->
+          let !(# ne, bit #) = WW.shr1_c# ex
+              !candidate = mul# r m
+              !nr = select# r candidate bit
+              !nm = sqr# m
+          in  loop nr nm ne (n - 1)
+  in  loop o b e (256 :: Word)
+{-# INLINE exp# #-}
 
 odd# :: (# Limb, Limb, Limb, Limb #) -> C.Choice
 odd# = WW.odd#
-{-# INLINE odd #-}
+{-# INLINE odd# #-}
 
 -- | Check if a 'Montgomery' value is odd.
 --
+--   Note that the comparison is performed in constant time, but we
+--   branch when converting to 'Bool'.
+--
 --   >>> odd 1
 --   True
 --   >>> odd 2
 --   False
 --   >>> Data.Word.Wider.odd (retr 3) -- parity is preserved
 --   True
-odd :: Montgomery -> Bool
-odd (Montgomery m) = C.decide (odd# m)
+odd_vartime :: Montgomery -> Bool
+odd_vartime (Montgomery m) = C.decide (odd# m)
 
 -- constant-time selection ----------------------------------------------------
 
